from typing import Hashable, ItemsView
from abc import ABC, abstractmethod


class EnumerableHandler(ABC):
    """EnumerableHandler is a base class for concrete enumerable object handlers.
    Such a handler must implement all abstract methods to allow handling of enumerable objects
    """

    @abstractmethod
    def copy_type(self):
        """return an empty instance of the object type associated with this handler
        """
        pass

    @abstractmethod
    def iterate(self, item: any) -> ItemsView[any, any]:
        """return a view on the key/value pairs of the given item

        Args:
            item (any): an object supported by this handler
        """
        pass

    @abstractmethod
    def put(self, target: any, key: any, value: any) -> None:
        """insert the given key & value into the target

        Args:
            target (any): the object to insert into
            key (any): the key
            value (any): the value
        """
        pass

    @abstractmethod
    def can_handle(self, item: any) -> bool:
        """return whether this handler supports the given item

        Args:
            item (any): the object to test

        Returns:
            bool: whether this handler supports the given item
        """
        pass

    @abstractmethod
    def contains_key(self, key: any, item: any) -> bool:
        """whether the given item contains the given key

        Args:
            key (any): the key to test
            item (any): the item to test against

        Returns:
            bool: whether the given item contains the given key
        """
        pass


class ListHandler(EnumerableHandler):

    def copy_type(self):
        return list()

    def iterate(self, item: list) -> ItemsView[int, any]:
        return enumerate(item)

    def put(self, target: list, key: int, value: any) -> None:
        if isinstance(key, int) and 0 <= key < len(target):
            target[key] = value
        else:
            target.append(value)

    def can_handle(self, item: any) -> bool:
        return isinstance(item, list)

    def contains_key(self, key: int, item: list) -> bool:
        if isinstance(key, int):
            return 0 <= key < len(item)
        return False


class DictHandler(EnumerableHandler):

    def copy_type(self) -> dict:
        return dict()

    def iterate(self, item: dict) -> ItemsView[Hashable, any]:
        return item.items()

    def put(self, target: dict, key: Hashable, value: any) -> None:
        target[key] = value

    def can_handle(self, item) -> bool:
        return isinstance(item, dict)

    def contains_key(self, key: Hashable, item: dict) -> bool:
        return key in item


_handlers: set[EnumerableHandler] = {ListHandler(), DictHandler()}


def get_handler(item):
    for handler in _handlers:
        if handler.can_handle(item):
            return handler
    raise Exception("No handler found for type %s" % type(item))



from typing import List, Union
import copy
import numpy as np
from decomply.enumerable import EnumerableHandler, ListHandler, DictHandler


class Decomply:
    """
    Decomply class allows unified processing of nested dictionaries (or any enumerable objects). Main method is decomply
    """

    def __init__(
        self,
        traverse: callable = lambda trace, item: True,
        apply: callable = lambda trace, item: item,
        delete: callable = lambda trace, item: str(trace[-1]).startswith("_"),
        handlers: set[EnumerableHandler] = None
    ):
        """
        Initialize the Decomply object

        Args:
        traverse (callable, optional): A callable function that determines whether
            to traverse further into nested objects during cascading. It takes
            two arguments: `trace` (list) - the current trace of keys, and `item`
            (any) - the current item being evaluated. Defaults to always return True.
        apply (callable, optional): A callable function that applies a transformation
            or operation on an item during cascading when `traverse` condition is not
            met. It takes two arguments: `trace` (list) - the current trace of keys,
            and `item` (any) - the current item being evaluated. Defaults to returning
            the item unchanged.
        delete (callable, optional): A callable function that determines whether to
            delete a key/value pair during cascading based on the key and value. It
            takes two arguments: `trace` (list) - the current trace, and `item` (any) -
            the current item associated with the trace. Defaults to deleting keys that
            start with an underscore (`key.startswith("_")`).
        handlers (set[EnumerableHandler], optional): A set of @EnumerableHandlers. 
            Such a handler must implement methods to allow handling of an enumerable 
            object. Users may write their own handlers and supply them here, to 
            extend support. In basekit, only list and dict is supported.
        """
        self.traverse = lambda trace, item: self._is_enumerable(
            item) and traverse(trace, item)
        self.apply = apply
        self.delete = delete

        self.handlers: set[EnumerableHandler] = {ListHandler(), DictHandler()}
        if handlers:
            self.handlers.update(handlers)

    def decomply(self, item: Union[dict, list, any], trace: List[Union[str, int, any]] = None, initial_check: bool = True) -> Union[dict, list, any]:
        """
        Decomply the supplied `item`

        Args:
            item (Union[dict, list, any]): The object to be decomplied. Supported is list and dict, but 
                using your own handlers, you may pass any object your handlers support
            trace (List[Union[str, int, any]], optional): A list representing the current trace of keys
                during traversing. Defaults to None.
            initial_check (bool, optional): Flag indicating whether to perform the initial check
                with the `traverse` condition defined in the constructor (`__init__` method).
                Defaults to True.

        Returns:
            Union[dict, list, any]: An object with the same structure as `item`, modified according to the
                configured `traverse`, `apply`, and `delete` behaviors.

        Notes:
            - The `traverse`, `apply`, and `delete` behaviors are configured during the initialization
            of the Decomply object (`__init__` method).
            - If `trace` is not provided (`None`), it defaults to an empty list

        """
        if isinstance(trace, type(None)):
            trace = list()
        if initial_check and not self.traverse(trace, item):
            return self.apply(item, copy.deepcopy(trace))
        return self._decomply(trace, item)

    def _decomply(self, trace: List[Union[str, int]], item: Union[dict, list, any]) -> Union[dict, list, any]:
        """
        Traverse the item's items in a depth-first search style (DFS)
        Apply if traverse evaluates to False.
        Build the output object iteratively.

        Returns:
            Union[dict, list, any]: see @decomply
        """
        _trace = copy.deepcopy(trace)  # prevent side effects
        handler = self._get_handler(item)
        out = handler.copy_type()
        for key, value in handler.iterate(item):
            _trace.append(key)
            if not self.delete(_trace, value):
                if self.traverse(_trace, value):
                    handler.put(out, key, self._decomply(_trace, value))
                else:
                    handler.put(out, key, self.apply(
                        copy.deepcopy(_trace), value))
            _trace.pop()
        return out

    def _get_handler(self, item) -> EnumerableHandler:
        for handler in self.handlers:
            if handler.can_handle(item):
                return handler
        raise Exception("No handler found for type %s" % type(item))

    def _is_enumerable(self, item: any) -> bool:
        return np.any([handler.can_handle(item) for handler in self.handlers])



from setuptools import setup, find_packages

setup(
    name='decomply',
    version='1.1.0',
    packages=find_packages(),
    description='decomply allows unified processing of nested dictionaries',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',  # for Markdown
    author='TrueKenji',
    url='https://github.com/truekenji/decomply',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    keywords=['Dictionary', 'Recursive', 'Traverse', 'Apply',
              'Decompose', 'Partition', 'Nested', 'Loop', 'List', 'JSON']
)



import unittest
from decomply import Decomply


class TestDecomply(unittest.TestCase):

    def test_decomply_example(self):
        def apply(_, item): return item + 1

        decomply_instance = Decomply(apply=apply)
        input_data = {
            "First Layer": {
                "Second layer": 3,
                "_Second layer 2nd element": 4
            }
        }
        expected_output = {
            "First Layer": {
                "Second layer": 4
            }
        }
        self.assertEqual(decomply_instance.decomply(input_data), expected_output)

    def test_decomply_example2(self):
        input_data = {
            "First Layer": {
                "Second layer": 3,
                "_Second layer 2nd element": 4,
                "Second layer 3rd element": [
                    42,
                    69
                ]
            },
            "First layer 2nd element": {
                "fuu": {
                    "fuu 1st entry": 6,
                    "fuu 2nd entry": 10,
                    "fuu 3rd entry": {
                        "I will be dropped": 0
                    }
                }
            }
        }
        expected_output = {
            "First Layer": {
                "Second layer": 4,
                "_Second layer 2nd element": 5,
                "Second layer 3rd element": [
                    42,
                    69,
                    666
                ]
            },
            "First layer 2nd element": {
                "fuu": {
                    "fuu 1st entry": 7,
                    "fuu 2nd entry": 11,
                    "fuu 3rd entry": {}
                }
            }
        }
        out = Decomply(
            traverse=lambda _, item: not isinstance(item, list),
            apply=lambda _, item: item +
            [666] if isinstance(item, list) else item + 1,
            delete=lambda trace, _: len(trace) > 3
        ).decomply(input_data)
        self.assertEqual(out, expected_output)


if __name__ == '__main__':
    unittest.main()
